// LED Curve Tracer
// Ed Nisley - KE4ANU - July 2012

#include <stdio.h>

//----------
// Pin assignments

const byte PIN_READ_LEDSUPPLY = 0;	// AI - LED supply voltage		blue
const byte PIN_READ_VDRAIN = 1;		// AI - drain voltage			red
const byte PIN_READ_VSOURCE = 2;	// AI - source voltage			orange
const byte PIN_READ_VGATE = 3;		// AI - VGS after filtering		violet

const byte PIN_SET_VGATE = 11;		// PWM - gate voltage			brown

const byte PIN_BUTTON1 = 8;			// DI - button to start tests	green
const byte PIN_BUTTON2 = 7;			// DI - button for options		yellow

const byte PIN_HEARTBEAT = 13;		// DO - Arduino LED
const byte PIN_SYNC = 2;			// DO - scope sync output

//----------
// Constants

const int MaxCurrent = 30;				// maximum LED current - mA
const int ISTEP = 5;					// LED current increment

const float Vcc = 4.96; 				// Arduino supply -- must be measured!

const float RSense = 10.500;			// current sense resistor

const float ITolerance = 0.0005;		// current setpoint tolerance

const float VGStep = 0.019;				// increment/decrement VGate = 5 V / 256

const byte PWM_Settle = 5;				// PWM settling time ms

#define TCCRxB 0x01						// Timer prescaler = 1:1 for 32 kHz PWM

#define MK_UL(fl,sc) ((unsigned long)((fl)*(sc)))
#define MK_U(fl,sc) ((unsigned int)((fl)*(sc)))

//----------
// Globals

float AVRef1V1;					// 1.1 V bandgap reference - calculated from Vcc

float VccLED;					// LED high-side supply

float VDrain;					// MOSFET terminal voltages
float VSource;
float VGate;

unsigned int TestNum = 1;

long unsigned long MillisNow;

//-- Read AI channel
//      averages several readings to improve noise performance
//		returns value in mV assuming VCC ref voltage

#define NUM_T_SAMPLES    10

float ReadAI(byte PinNum) {

  word RawAverage;

  digitalWrite(PIN_SYNC,HIGH);                // scope sync

  RawAverage = analogRead(PinNum);            // prime the averaging pump

  for (int i=2; i <= NUM_T_SAMPLES; i++) {
    RawAverage += (word)analogRead(PinNum);
  }

  digitalWrite(PIN_SYNC,LOW);

  RawAverage /= NUM_T_SAMPLES;

  return Vcc * (float)RawAverage / 1024.0;

}

//-- Set PWM output

void SetPWMVoltage(byte PinNum,float PWMVolt) {

byte PWM;

  PWM = (byte)(PWMVolt / Vcc * 255.0);

  analogWrite(PinNum,PWM);
  delay(PWM_Settle);

}

//-- Set VGS to produce desired LED current
//		bails out if VDS drops below a sensible value

void SetLEDCurrent(float ITarget) {

  float ISense;				// measured current
  float VGateSet;			// output voltage setpoint
  float IError;				// (actual - desired) current

  VGate = ReadAI(PIN_READ_VGATE);					// get gate voltage
  VGateSet = VGate;									//  because input may not match output

  do {

	VSource = ReadAI(PIN_READ_VSOURCE);
	ISense = VSource / RSense;						// get LED current

//	printf("\r\nITarget: %lu mA",MK_UL(ITarget,1000.0));
	IError = ISense - ITarget;

//	printf("\r\nISense: %d mA VGateSet: %d mV VGate %d IError %d mA",
//		   MK_U(ISense,1000.0),
//		   MK_U(VGateSet,1000.0),
//		   MK_U(VGate,1000.0),
//		   MK_U(IError,1000.0));

	if (IError < -ITolerance) {
	  VGateSet += VGStep;
//	  Serial.print('+');
	}
	else if (IError > ITolerance) {
	  VGateSet -= VGStep;
//	  Serial.print('-');
	}

	VGateSet = constrain(VGateSet,0.0,Vcc);
	SetPWMVoltage(PIN_SET_VGATE,VGateSet);

	VDrain = ReadAI(PIN_READ_VDRAIN);		// sample these for the main loop
	VGate = ReadAI(PIN_READ_VGATE);
	VccLED = ReadAI(PIN_READ_LEDSUPPLY);

	if ((VDrain - VSource) < 0.020) {			// bail if VDS gets too low
	  printf("# VDS=%d too low, bailing\r\n",MK_U(VDrain - VSource,1000.0));
	  break;
	}

  } while (abs(IError) > ITolerance);


//	Serial.println(" Done");
}

//-- compute actual 1.1 V bandgap reference based on known VCC = AVcc (more or less)
//		adapted from http://code.google.com/p/tinkerit/wiki/SecretVoltmeter

float ReadBandGap(void) {

  word ADCBits;
  float VBandGap;

  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);	// select 1.1 V input
  delay(2); // Wait for Vref to settle

  ADCSRA |= _BV(ADSC);										// Convert
  while (bit_is_set(ADCSRA,ADSC));

  ADCBits = ADCL;
  ADCBits |= ADCH<<8;

  VBandGap = Vcc * (float)ADCBits / 1024.0;
  return VBandGap;
}


//-- Print message, wait for a given button press

void WaitButton(int Button,char *pMsg) {
  printf("# %s",pMsg);
  while(HIGH == digitalRead(Button)) {
    delay(100);
    digitalWrite(PIN_HEARTBEAT,!digitalRead(PIN_HEARTBEAT));
  }

  delay(50);				// wait for bounce to settle
  digitalWrite(PIN_HEARTBEAT,LOW);
}


//-- Helper routine for printf()

int s_putc(char c, FILE *t) {
  Serial.write(c);
}


//------------------
// Set things up

void setup() {
  pinMode(PIN_HEARTBEAT,OUTPUT);
  digitalWrite(PIN_HEARTBEAT,LOW);	// show we arrived

  pinMode(PIN_SYNC,OUTPUT);
  digitalWrite(PIN_SYNC,LOW);		// show we arrived

  TCCR1B = TCCRxB;					// set frequency for PWM 9 & 10
  TCCR2B = TCCRxB;					// set frequency for PWM 3 & 11

  pinMode(PIN_SET_VGATE,OUTPUT);
  analogWrite(PIN_SET_VGATE,0);		// force gate voltage = 0

  pinMode(PIN_BUTTON1,INPUT_PULLUP);	// use internal pullup for buttons
  pinMode(PIN_BUTTON2,INPUT_PULLUP);

  Serial.begin(57600);
  fdevopen(&s_putc,0);				// set up serial output for printf()

  printf("# LED Curve Tracer\r\n# Ed Nisley - KE4ZNU - March 2013\r\n");

  VccLED = ReadAI(PIN_READ_LEDSUPPLY);
  printf("# VCC at LED: %d mV\r\n",MK_U(VccLED,1000.0));

  AVRef1V1 = ReadBandGap();			// compute actual bandgap reference voltage
  printf("# Bandgap reference voltage: %lu mV\r\n",MK_UL(AVRef1V1,1000.0));

}

//------------------
// Run the test loop

void loop() {

  Serial.println('\n');						// blank line for Gnuplot indexing

  WaitButton(PIN_BUTTON1,"Insert LED, press button 1 to start...\r\n");
  printf("# INOM\tILED\tVccLED\tVD\tVLED\tVG\tVS\tVGS\tVDS\t<--- LED %d\r\n",TestNum++);
  digitalWrite(PIN_HEARTBEAT,LOW);

  for (int ILED=0; ILED <= MaxCurrent; ILED+=ISTEP) {
	SetLEDCurrent(((float)ILED)/1000.0);
	printf("%d\t%lu\t%d\t%d\t%d\t%d\t%d\t%d\t%d\r\n",
		    ILED,
			MK_UL(VSource / RSense,1.0e6),
			MK_U(VccLED,1000.0),
			MK_U(VDrain,1000.0),
			MK_U(VccLED - VDrain,1000.0),
			MK_U(VGate,1000.0),
			MK_U(VSource,1000.0),
			MK_U(VGate - VSource,1000),
			MK_U(VDrain - VSource,1000.0)
		  );
  }

  SetPWMVoltage(PIN_SET_VGATE,0.0);

}

